package dev.inmo.postssystem.core.publishing

import dev.inmo.postssystem.core.post.PostId
import dev.inmo.postssystem.core.publishing.repos.PublishingKeysRepo
import kotlinx.coroutines.channels.BroadcastChannel
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.asFlow

typealias TriggerControlKey = String

interface ReadPublishingRegistrar {
    suspend fun getPostIdByTriggerControlKey(
        key: TriggerControlKey
    ): PostId?
}

interface WritePublishingRegistrar {
    val unregisteredKeysFlow: Flow<TriggerControlKey>

    suspend fun registerTriggerForPost(
        key: TriggerControlKey,
        postId: PostId
    ): Boolean
}

interface PublishingRegistrar : ReadPublishingRegistrar, WritePublishingRegistrar

class BusinessPublishingRegistrar(
    private val repo: PublishingKeysRepo
) : PublishingRegistrar {
    private val unregisteredKeysChannel: BroadcastChannel<TriggerControlKey> = BroadcastChannel(Channel.BUFFERED)
    override val unregisteredKeysFlow: Flow<TriggerControlKey> = unregisteredKeysChannel.asFlow()

    override suspend fun getPostIdByTriggerControlKey(
        key: TriggerControlKey
    ): PostId? = repo.getPostIdByTriggerControlKey(key)

    override suspend fun registerTriggerForPost(
        key: TriggerControlKey,
        postId: PostId
    ): Boolean = repo.getTriggerControlKeyByPostId(
        postId
    ).let { previousKey ->
        repo.setPostTriggerControlKey(postId, key).also { inserted ->
            if (inserted && previousKey != null) {
                unregisteredKeysChannel.send(previousKey)
            }
        }
    }
}

